/***************************************************************************************
*
* Block_FIR.ASM
*
* Accompaniment for:
* Guide for Porting Code from SHARC DSPs to TigerSHARC Processors (EE-241)
*
* Written by Andrew Caldwell and Maikel Kokaly-Bannourah 
*	
* June 2004 rev 1.0 
*
* This modified version of the Block FIR code shows a one-to-one conversion (optimization 3)
* of the Block FIR SHARC code to TigerSHARC. 
*
***************************************************************************************/

/***********************************************************************

This is an example of a C callable floating point block FIR.
The example is for 64 taps with 128 samples and is provided to show
how the techniques detailed in the accompanying application note can be
applied to this example to convert it for operation on the TigerSHARC
family of processors.

Some of the macros for saving and restoring to and from the stack
have been added simply to show how the TigerSHARC macros provided with
the accompanying application note can be implemented without much
alteration of the code.

Optimizations performed in this example are:

1) Re-ordering of instructions to make more parallel, this will make much
   better use of the 128-bit instruction line and will reduce a number of stalls
   as well as decrease the number of instruction lines in the example.
   
2) Optimizations performed in the previous examples are included here.
   This includes using quad access for both delay line reads and coefficient
   reads. The dline reads are performed using distributed quad loads through
   the DAB. Also some removal of some instructions were performed after the
   initial conversion to the TigerSHARC. These did not affect functionality
   but were simply not efficient and redundant.
   
There is still a stall in the inner loop. This is due to the fact that there
is a single stall due to compute block load dependency. This can be overcome
by loop unrolling. The doubling up on the contents of the loop so 8 dline samples 
and 8 coefficients are loaded every iteration and the compute instructions interleaved so
there is always two cycles between the load of the registers and the use of
the registers in the multiply and accumulate instructions. This would
double the cycle count for the loop but half the amount of times the innrer
loop is executed and remove the stall. For a fully optimized example refer to
the code examples included with the VisualDSP++ instalation.
   
************************************************************************/

#ifdef __ADSPTS201__
	#include <defts201.h>
#endif

#ifdef __ADSPTS101__
	#include <defts101.h>
#endif

#include "asm_sprt_TS.h"

.global _block_fir;

/* program memory code */
.section program;

_block_fir:
	leaf_entry;

/****************************************************
*	Save various registers to stack
*	This is simply added to show the use of the macros
*	In this version we use both J and K stacks
*****************************************************/	
	/* save all register block to stack */
	save_reg;
	
//1	/* save some other registers to stack */
	puts_jstack = j0;	puts_kstack = k0;;
	puts_jstack = j5; 	puts_kstack = k5;;



/****************************************************
* 	Retrieve passed arguments
*	The 4 pointers are passed through J7:4
*	The number of samples and number of taps
*	are integer values passed on the stack
*	j4 = input
*	j5 = dline
*	j6 = coeffs
*	j7 = output
*	reads(1) = samples
*	reads(2) = TAPS
*****************************************************/	
	
	xr17 = reads(1);	// N number of samples
	xr16 = reads(2);	// Number of TAPS 
	
	/* dline pointer CB setup */
	/* second parameter passed stired in r8 */
	j0 = j5;;
	jb0 = j0;;
	jl0 = xr16;;

/* Create new pointer in dline for storing new input */
	j1 = j0;;
	jb1 = j0;;
	jl1 = xr16;;
	
	/* input samples buffer setup */
	/* first parameter passed stored in r4 */
	j5 = j4;;
	
	/* coeffs pointer CB setup */
	/* third parameter passed stored in r12 */
	k0 = j6;;
	kb0 = k0;;
	kl0 = xr16;;
	
	/* output buffer setup */
	k5 = j7;;
	
/* Setup modify registers for arrays, calculate loop counts, prime the DAB and load first samples and coefficients */ 
	j12 = 0; r3:0 = DAB q[j0+=0]; xr16 = lshift r16 by -1;;
	j13 = 1;;
	j14 = -1; r3:0 = DAB q[j0+=4];;
	j15 = 2;;
	xr18 = 4;;
	k13 = 2;r3:0 = DAB q[j0+=4];;
	xr18 = r16 - r18; LC0 = xr17;;
	k14 = 1; xr0 = [j5+=j13];;
	xr18 = lshift r18 by -1; yxr5:4 = CB q[k0+=4];;	

	
		
main_fir:	CB [j1+=j14] = xr0; r9 = 0;;			/* transfer new sample to delay line and reset second accum register */
			LC1 = xr18;;							/* Set inner loop counter */
			xfr8 = r0*r4; yfr8 = r2*r4; r13 = 0;;	/* samples * coeffs, resets second MAC register */
			xfr12 = r1*r5; yfr12 = r3*r5; yxr5:4 = CB q[k0+=4]; r3:0 = DAB q[j0+=4];;	/* samples * coeffs, read 4 samples, read 4 coeffs */

  
		/* FIR loop */
				
macs:   nop;;
		xfr12 = r0*r4; yfr12 = r2*r4; fr8 = r8+r12;;		/* samples * coeffs, accum */
		xfr13 = r1*r5; yfr13 = r3*r5; fr9 = r9+r13;;		/* samples * coeffs, accum */
		if NLC1E, jump macs; yxr5:4 = CB q[k0+=4]; r3:0 = DAB q[j0+=4];; /* loop, load 4 coeffs, load 4 samples*/
        
		/* samples * coeffs, accum, reset pointer to new dline[0] */
		xfr12 = r0*r4; yfr12 = r2*r4; fr8 = r8+r12; j0 = j0 - 5 (cb);;

		/* samples * coeffs, accum, first prime of DAB */
		xfr13 = r1*r5; yfr13 = r3*r5; fr9 = r9+r13; r3:0 = DAB q[j0+=0];;
        
		/* accum, read new sample into s10 */
		fr8 = r8+r12; yr10 = [j5+=j13];;

		/* accum second results and second prime of DAB, move Y accum to X */
		fr9 = r9+r13; r3:0 = DAB q[j0+=4]; xr12 = yr8;;
        
		/* move Y accum into X register file */
		xr13 = yr9;;

		/* Accum results of Y and X, load 4 coefficients*/
		fr8 = r8+r12; yxr5:4 = CB q[k0+=4];;
		
		/* Accum results of Y and X */
		fr9 = r9+r13;;
		
		/* Final accum, load first 4 samples*/
		fr8 = r8+r9; r3:0 = DAB q[j0+=4];;		
		/* prime the DAB again */
		
		/* write result to OUTPUT[i], load new sample to X as dline has not been updated until */
if NLC0E, jump main_fir; [k5+=k14] = xr8; xr0 = yr10;;

	
	/* restore registers from J and K stack */
	k5 = gets_kstack(1) j5 = gets_jstack(1);
	k0 = gets_kstack(2) j0 = gets_jstack(2);
	
	/* modify J and K stack pointers */
	alter_jstack(2) alter_kstack(2);
	
	/* restore register block from stack */
	restore_reg;
		
_block_fir.END:	leaf_exit;
